/* Generic C includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>

/* Structures, protos and erreurs of the lib */
#include "libregionMap.h"

/**********************************************************************
 *
 * Computes the partial probabilities for a region
 *
 **********************************************************************/

int	
rMap_computeMap3DProbas (REGION_MAP*		rm,
			 RMAP_3D_PROTO_SPACE*	ps,
			 RMAP_FUSION_TYPE	fusionType)
{
  int		i,j,k;
  RMAP_REGION*	region;
  RMAP_REGION*	prototype;
  RMAP_REGION*	closestPrototype;
  double	distance, minDist;
  double	sumProbas;

  /* Check that the PROTO_SPACE 'ps' is in the proper state */
  if (ps->state != PROBAS_PROTOS_OK) {
    printf ("Partial probabilities of the PROTO_SPACE not computed\n");
    return rMapError (ERR_RMAP_BAD_PARAMETER);
  }
  if (rm->lastCellStates != RMAP_ATTRIBUTES_COMPUTED) {
    printf ("Attributes of the REGION_MAP not computed\n");
    return rMapError (ERR_RMAP_BAD_PARAMETER);
  }

  for (i = rm->iMin; i <= rm->iMax; i++) {
    for (j = rm->jMin; j <= rm->jMax; j++) {
      region = rm->regions + (rm->nbcol * i) + j;
      if (region->updateState == RMAP_REGION_NEWLY_PERCEIVED) {
	/* Normalizes the attributes */
	rMap_normalize3DAttributes (region, ps, ps->norm);
	/* Find the closest prototype */
	minDist = FLT_MAX;
	closestPrototype = NULL;
	for (k = 0; k < ps->nbProtos; k++) {
	  prototype = ps->prototypes + k;
	  distance = 
	    rMap_compute3DProtoDistance (region, prototype, ps->attributeMask);
	  if (distance < minDist) {
	    minDist = distance;
	    closestPrototype = prototype;
	  }
	}
	if (closestPrototype == NULL) {
	  printf ("Smth wrong, impossible to find the closest prototype !\n");
	  return rMapError (ERR_RMAP_IMPOSSIBLE_CASE);
	}
	else {			/* Sets the partial probas of the region as the
				   ones of the closest prototype */
	  switch (fusionType) {
	  case RMAP_FUSION_OVERRIDE:
	    memcpy (region->infos3d.newProbas, 
		    closestPrototype->infos3d.mergedProbas,
		    NB_3D_CLASSES * sizeof(double));
	    memcpy (region->infos3d.mergedProbas, 
		    closestPrototype->infos3d.mergedProbas,
		    NB_3D_CLASSES * sizeof(double));
	    region->infos3d.probaState = RMAP_PROBAS_COMPUTED;
	    break;
	  case RMAP_FUSION_MERGE:
	    memcpy (region->infos3d.newProbas, 
		    closestPrototype->infos3d.mergedProbas,
		    NB_3D_CLASSES * sizeof(double));
	    sumProbas = 0.0;
	    for (k = 0; k < NB_3D_CLASSES; k++) {
	      sumProbas += (closestPrototype->infos3d.mergedProbas[k] 
			    * region->infos3d.mergedProbas[k]);
	    }
	    for (k = 0; k < NB_3D_CLASSES; k++) {
	      region->infos3d.mergedProbas[k] 
		= (closestPrototype->infos3d.mergedProbas[k] 
		   * region->infos3d.mergedProbas[k]) / sumProbas;
	    }
	    region->infos3d.probaState = RMAP_PROBAS_COMPUTED;
	    break;
	  default:
	    printf ("can not handle fusion type %d\n", fusionType);
	    return rMapError (ERR_RMAP_IMPOSSIBLE_CASE);
/* 	    break; Unreached */
	  }
	}
      }
    }
  }

  /* Updates some attributes of the regionMap */
  memcpy (rm->classes3d, ps->classes, NB_3D_CLASSES * sizeof(char));
  rm->lastCellStates = RMAP_PROBAS_COMPUTED;

  return (RMAP_OK);
}

/**********************************************************************
 *
 * Simple version to label a Map (set its class to the highest partial
 * probability) 
 *
 **********************************************************************/

int	
rMap_computeMap3DLabel (REGION_MAP*		rm)

{
  int		i,j;
  RMAP_REGION*	region;
  double	probaMax;
  RMAP_3D_CLASS	highestProbaClass;
  RMAP_3D_CLASS	k;

  if (rm->lastCellStates != RMAP_PROBAS_COMPUTED) {
    printf ("Partial probabilities of the REGION_MAP not computed\n");
    return rMapError (ERR_RMAP_BAD_PARAMETER);
  }

  for (i = rm->iMin; i <= rm->iMax; i++) {
    for (j = rm->jMin; j <= rm->jMax; j++) {
      region = rm->regions + (rm->nbcol * i) + j;
      if (region->infos3d.probaState == RMAP_PROBAS_COMPUTED) {
	probaMax = 0.0;
	for (k = 0; k < NB_3D_CLASSES; k++) {
	  if (rm->classes3d[k] != 0) {
	    if (region->infos3d.mergedProbas[k] > probaMax) {
	      probaMax = region->infos3d.mergedProbas[k];
	      highestProbaClass = k;
	    }
	  }
	}
	region->infos3d.label = highestProbaClass;
	region->infos3d.classificationState = SYSTEM_CLASSIFIED;
      }
    }
  }

  /* Updates some attributes of the regionMap */
  rm->lastCellStates = RMAP_AUTOMATICALLY_LABELLED;

  return (RMAP_OK);
}
