/* Generic C includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

/* Structures, protos and erreurs of the lib */
#include "libregionMap.h"

#ifndef MAX
#define MAX(a,b) ((a) > (b) ? (a) : (b))
#endif

#ifndef MIN
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif

 
/**********************************************************************
 *
 * Computes the min/max coordinates of updated cells (this function should
 * actually never be called, as the bounds informations are properly updated
 * when fusing 3D images in the regionMap).
 *
 **********************************************************************/

int
rMap_computeBounds (REGION_MAP*	regionMap)

{
  RMAP_REGION*	region;
  int		i,j;

  region = regionMap->regions;
  for (i = 0; i < regionMap->nblig; i++) {
    for (j = 0; j < regionMap->nbcol; j++) {
      if (region->updateState != RMAP_REGION_NEVER_PERCEIVED) {
	regionMap->iMin = MIN (regionMap->iMin, i);
	regionMap->iMax = MAX (regionMap->iMax, i);
	regionMap->jMin = MIN (regionMap->jMin, j);
	regionMap->jMax = MAX (regionMap->jMax, j);
      }
      region++;
    }
  }

  regionMap->bounds = RMAP_BOUNDS_UPDATED;
  return (RMAP_OK);
}

/*
 * Simple function called by the macro "tbError"
 * Prints infos on the error that occured
 */

int
rMap_PrintError(int num, 
		char *errorCode, 
		char *fileName, 
		int lineNumber)
{
  printf ("ErrorCode: %s\n", errorCode);
  printf ("File: %s\n", fileName);
  printf ("Line: %d\n", lineNumber);
  return (num);
}

/**********************************************************************
 *
 * Creates a new REGION_MAP whose size strictly fits the bounding areas of the
 * given REGION_MAP
 *
 **********************************************************************/

REGION_MAP*
rMap_autocrop (REGION_MAP*     rMap)
{
  REGION_MAP*	new_rMap;
  RMAP_REGION*	new_region;

  int		nblig, nbcol;

  int		i_min, j_min;
  int		new_nblig, new_nbcol;
  double	xOrigin, yOrigin, zOrigin;
  
  int		i, j;
  
  nblig = rMap->nblig;
  nbcol = rMap->nbcol;
  
  /* Compute the bounds of the rMap */
  rMap_computeBounds (rMap);
  
  /* Compute various parameters that specifies the new one */
  i_min = rMap->iMin;
  j_min = rMap->jMin;
  new_nblig = rMap->iMax - rMap->iMin + 1;
  new_nbcol = rMap->jMax - rMap->jMin + 1;

  /* Finds the metric coordinates of the first region */
  xOrigin = i_min * rMap->scale + rMap->origin2corner.euler.euler[3];
  yOrigin = j_min * rMap->scale + rMap->origin2corner.euler.euler[4];
  zOrigin = rMap->origin2corner.euler.euler[5];
  
  /* Creates and initialize the new rMap */
  new_rMap = rMap_createMap ();
  if (new_rMap == NULL) {
    rMapError(ERR_RMAP_ALLOC_REGION_MAP);
    return (NULL);
  }
  
  if (rMap_initMap (new_rMap, rMap->scale,
		    new_nblig, new_nbcol, 
		    xOrigin, yOrigin, zOrigin,
		    rMap->robotRadius) != RMAP_OK) {
    rMapError (ERR_RMAP_ALLOC_REGION_MAP);
    return (NULL);
  }
  
  /* Copy some general update informations */
  new_rMap->lastCellStates = rMap->lastCellStates;
  new_rMap->bounds = rMap->bounds;
  new_rMap->geoOrientation = rMap->geoOrientation;
  new_rMap->nbUpdates = rMap->nbUpdates;
  new_rMap->nbNewCells = rMap->nbNewCells;
  new_rMap->nbUpdatedCells = rMap->nbUpdatedCells;
  memcpy (&(new_rMap->geodesicOrigin), &(rMap->geodesicOrigin), 
	  sizeof (DTM_GEODESIC_POINT));
  for (i = 0; i < RMAP_NB_POSITIONS_MAX; i++) 
    new_rMap->positions[i] = rMap->positions[i];
  new_rMap->nbPointsMin = rMap->nbPointsMin;
  memcpy (new_rMap->classes3d, rMap->classes3d, 
	  NB_3D_CLASSES * sizeof (int));
  memcpy (new_rMap->classesTexture, rMap->classesTexture, 
	  NB_3D_CLASSES * sizeof (int));

  /* Now loop on the regions and copy them */
  new_region = new_rMap->regions;
  for (i = 0; i < new_rMap->nblig; i++) {
    for (j = 0; j < new_rMap->nbcol; j++) {
      memcpy ((RMAP_REGION*) new_region, 
	      (RMAP_REGION*) (rMap->regions + nbcol * (i + i_min) + (j + j_min)),
	      sizeof (RMAP_REGION));
      new_region++;
    }
  }
  
  
  return (new_rMap);
}

/**********************************************************************
 *
 * Creates a new REGION_MAP which is subsampled by a intiger factor
 *
 **********************************************************************/

REGION_MAP*
rMap_subSample (REGION_MAP*     rMap,
		int		factor)
{
  REGION_MAP*	new_rMap;
  RMAP_REGION*	new_region;
  RMAP_REGION*	region;
  RMAP_REGION*	mostObstacleMergedRegion;

  int		nblig, nbcol;
  int		new_nblig, new_nbcol;
  double	xOrigin, yOrigin, zOrigin;
  
  int		i, j, k, l;
  double	mergedProbaObstMax;
  
  nblig = rMap->nblig;
  nbcol = rMap->nbcol;

  new_nblig = nblig / factor;
  new_nbcol = nbcol / factor;

  /* Creates and initialize the new rMap */
  new_rMap = rMap_createMap ();
  if (new_rMap == NULL) {
    rMapError(ERR_RMAP_ALLOC_REGION_MAP);
    return (NULL);
  }
  
  /* Finds the metric coordinates of the first region */
  xOrigin = rMap->origin2corner.euler.euler[3];
  yOrigin = rMap->origin2corner.euler.euler[4];
  zOrigin = rMap->origin2corner.euler.euler[5];

  if (rMap_initMap (new_rMap, rMap->scale * factor,
		    new_nblig, new_nbcol, 
		    xOrigin, yOrigin, zOrigin,
		    rMap->robotRadius) != RMAP_OK) {
    rMapError (ERR_RMAP_ALLOC_REGION_MAP);
    return (NULL);
  }

  /* Now loop on the regions and copy them */
  new_region = new_rMap->regions;
  for (i = 0; i < new_nblig; i++) {
    for (j = 0; j < new_nbcol; j++) {
      region 
	= rMap->regions + (nbcol * i * factor)  + (j * factor);
      mostObstacleMergedRegion = region;
      mergedProbaObstMax = 0.0;
      for (k = 0; k < factor; k++) {
	for (l = 0; l < factor; l++) {
	  region 
	    = rMap->regions + (nbcol * ((i * factor) + k))  + ((j * factor) + l);
	  if (region->infos3d.mergedProbas[RMAP_OBSTACLE] > mergedProbaObstMax) {
	    mergedProbaObstMax = region->infos3d.mergedProbas[RMAP_OBSTACLE];
	    mostObstacleMergedRegion = region;
	  }
	}
      }
      memcpy ((RMAP_REGION*) new_region, 
	      (RMAP_REGION*) mostObstacleMergedRegion,
	      sizeof (RMAP_REGION));
      new_region++;
    }
  }

  /* Computes some general attributes */
  rMap_computeBounds (new_rMap);
  new_rMap->lastCellStates = rMap->lastCellStates;
  new_rMap->nbUpdates = rMap->nbUpdates;
  
  new_region = new_rMap->regions;
  for (i = 0; i < new_nblig; i++) {
    for (j = 0; j < new_nbcol; j++) {
      if (new_region->updateState != RMAP_REGION_NEVER_PERCEIVED) 
	new_rMap->nbUpdatedCells++;
      if (new_region->lastDate == new_rMap->nbUpdates) {
	new_rMap->nbNewCells += 1;
	new_rMap->last_iMin = MIN (new_rMap->last_iMin, i);
	new_rMap->last_iMax = MAX (new_rMap->last_iMax, i);
	new_rMap->last_jMin = MIN (new_rMap->last_jMin, j);
	new_rMap->last_jMax = MAX (new_rMap->last_jMax, j);
      }
      new_region++;
    }
  }

  for (i = 0; i < RMAP_NB_POSITIONS_MAX; i++) 
    new_rMap->positions[i] = rMap->positions[i];
  
  return (new_rMap);
}

/**************************************************************
 *
 * Simply returns the string that correspond to a 3D class
 *
 **************************************************************/
char* 
rMap_return3dClassString (RMAP_3D_CLASS i)
{
  char*  classString;
  switch (i) {
  case NO_3D_CLASS: classString = "NO_3D_CLASS"; break;
  case RMAP_FLAT: classString = "RMAP_FLAT"; break;
  case RMAP_OBSTACLE: classString = "RMAP_OBSTACLE"; break;
  case RMAP_ROUGH: classString = "RMAP_ROUGH"; break;
  case RMAP_SLOPE: classString = "RMAP_SLOPE"; break;
  default: classString = "Undefined !!!"; break;
  }
  return (classString);
}

/**************************************************************
 *
 * Simply returns the string that correspond to a visual class
 *
 **************************************************************/
char* 
rMap_returnTextureClassString (RMAP_TEXTURE_CLASS i)
{
  char*  classString;
  switch (i) {
  case NO_TEXTURE_CLASS: classString = "NO_TEXTURE_CLASS"; break;
  case GRASS: classString = "GRASS"; break;
  case CONCRETE: classString = "CONCRETE"; break;
  case GRAVEL: classString = "GRAVEL"; break;
  case TARMAC: classString = "TARMAC"; break;
  default: classString = "Undefined !!!"; break;
  }
  return (classString);
}

/**************************************************************
 *
 * Simply returns the string that correspond to a 3D attribute
 *
 **************************************************************/
char* 
rMap_return3dAttributeString (RMAP_3D_ATTRIBUTES i)
{
  char*  classString;
  switch (i) {
  case NB_PTS: classString = "NB_PTS"; break;
  case MEAN_H_DISTANCE: classString = "MEAN_H_DISTANCE"; break;
  case MIN_Z: classString = "MIN_Z"; break;
  case MAX_Z: classString = "MAX_Z"; break;
  case DELTA_Z: classString = "DELTA_Z"; break;
  case MEAN_Z: classString = "MEAN_Z"; break;
  case VAR_Z: classString = "VAR_Z"; break;
  case NX: classString = "NX"; break;
  case NY: classString = "NY"; break;
  case NZ: classString = "NZ"; break;
  case VAR_NX: classString = "VAR_NX"; break;
  case VAR_NY: classString = "VAR_NY"; break;
  case VAR_NZ: classString = "VAR_NZ"; break;
  case NORM_VAR_N: classString = "NORM_VAR_N"; break;
  default: classString = "Undefined !!!"; break;
  }
  return (classString);
}

/**************************************************************
 *
 * Simply returns the string that correspond to a texture attribute
 *
 **************************************************************/
char* 
rMap_returnTextureAttributeString (RMAP_TEXTURE_ATTRIBUTES i)
{
  char*  classString;
  switch (i) {
  case ENERGY: classString = "ENERGY"; break;
  default: classString = "Undefined !!!"; break;
  }
  return (classString);
}


/**************************************************************
 *
 * Updates some region and regionMap parameters when a region is updated for 
 * first time.
 *
 **************************************************************/
void
rMap_updateStates (REGION_MAP*	regionMap,
		   RMAP_REGION*	region,
		   int		iRegion,
		   int		jRegion,
		   int		currentDate)
  
{
  region->updateState = RMAP_REGION_NEWLY_PERCEIVED;
  region->classificationType[CLASSIFICATION_3D] = 1;
  region->nbUpdates += 1;
  if (region->lastDate == 0) /* First time we perceive it */
    region->firstDate = currentDate;
  region->lastDate = currentDate;
  regionMap->nbNewCells += 1;
  regionMap->last_iMin = MIN (regionMap->last_iMin, iRegion);
  regionMap->last_iMax = MAX (regionMap->last_iMax, iRegion);
  regionMap->last_jMin = MIN (regionMap->last_jMin, jRegion);
  regionMap->last_jMax = MAX (regionMap->last_jMax, jRegion);
  return;
}

/**************************************************************
 *
 * Update the various global informations` on the regionMap after some of its
 * regions have been updated
 *
 **************************************************************/
void
rMap_endUpdate (REGION_MAP*	regionMap)

{
  regionMap->iMin = MIN (regionMap->last_iMin, regionMap->iMin);
  regionMap->iMax = MAX (regionMap->last_iMax, regionMap->iMax);
  regionMap->jMin = MIN (regionMap->last_jMin, regionMap->jMin);
  regionMap->jMax = MAX (regionMap->last_jMax, regionMap->jMax);
  regionMap->nbUpdates += 1;
  regionMap->nbUpdatedCells += regionMap->nbNewCells;
  regionMap->bounds = RMAP_BOUNDS_UPDATED;
  regionMap->lastCellStates = RMAP_ATTRIBUTES_COMPUTED;
  return;
}

/**********************************************************************
 *
 * Sets the geodesic reference of the regionMap (of the top-left corner of the
 * pixel (0,0) of the DTM.
 *
 **********************************************************************/

int 
rMap_setUTMGeodesicReference (REGION_MAP*	 rm,
			     int                 consideredPoint,
			     dtm_geo_orientation orientation,
			     double              easting,
			     double              northing,
			     double              alt,
			     double              nextEast)
{
  if ((orientation < 1) || (orientation > 4))
    return rMapError (ERR_RMAP_BAD_PARAMETER);

  if (consideredPoint != 0) {
    /* The provided geodetic coordinates are provided for the frame origin of
       the REGIONMAP */  
    switch (orientation) {
    case 1:			/* X = North */
      rm->geodesicOrigin.easting = easting - rm->origin2corner.euler.euler[4];
      rm->geodesicOrigin.northing = northing + rm->origin2corner.euler.euler[3];
      break;
    case 2:			/* X = West */
      rm->geodesicOrigin.easting = easting - rm->origin2corner.euler.euler[3];
      rm->geodesicOrigin.northing = northing - rm->origin2corner.euler.euler[4];
      break;
    case 3:			/* X = South */
      rm->geodesicOrigin.easting = easting + rm->origin2corner.euler.euler[4];
      rm->geodesicOrigin.northing = northing - rm->origin2corner.euler.euler[3];
      break;
    case 4:			/* X = East (should usually be the chosen, to
				   fit the ENU standard) */
      rm->geodesicOrigin.easting = easting + rm->origin2corner.euler.euler[3];
      rm->geodesicOrigin.northing = northing + rm->origin2corner.euler.euler[4];
      break;
    default:
      break; 
    }
  } else {
    /* The provided geodetic coordinates are the ones of the "top-left" pixel of
       the REGIONMAP, i.e. indices (0,0) */ 
    rm->geodesicOrigin.easting = easting;
    rm->geodesicOrigin.northing = northing;
  }
  rm->geodesicOrigin.alt = alt;
  rm->geoOrientation = orientation;
  rm->geodesicOrigin.nextEastMeridian = DEG_TO_RAD(nextEast);
  
  /* Computes the corresponding LLA coordinates */
  dtm_wgs84utm2lla (rm->geodesicOrigin.easting,
		    rm->geodesicOrigin.northing, 
		    rm->geodesicOrigin.nextEastMeridian,
		    &(rm->geodesicOrigin.lat), 
		    &(rm->geodesicOrigin.lon));

  return (RMAP_OK);
}

/**********************************************************************
 *
 * Computes the growedProbas
 *
 ***********************************************************************/

int
rMap_computeGrowProbas (REGION_MAP*	rm)

{
  RMAP_REGION*	region;
  int		i,j, k, l;
  double	maxObstacleProba;
  int		size = rm->robotCellRadius;

  for (i = MAX (rm->last_iMin - size, size); 
       i < MIN ((rm->nblig - size - 1), rm->last_iMax + size); 
       i++) {
    for (j = MAX (rm->last_jMin - size, size);
	 j < MIN ((rm->nbcol - size - 1), rm->last_jMax + size); 
	 j++) {

      maxObstacleProba = 0.0;
      for (k = i - size; k <= i + size; k++) {
	for (l = j - size; l <= j + size; l++) {
	  region = &(rm->regions[k * rm->nbcol + l]);
	  if (sqrt ((double) (k - i)*(k - i) + (l - j)*(l - j)) 
	      <= (double) (size)) 
	    maxObstacleProba = MAX (maxObstacleProba,
				    region->infos3d.mergedProbas[RMAP_OBSTACLE]);
	}
      }
      region = &(rm->regions[i * rm->nbcol + j]);
      if (maxObstacleProba > region->infos3d.mergedProbas[RMAP_OBSTACLE]) {
	region->infos3d.growedProbas[RMAP_OBSTACLE] = maxObstacleProba;
	region->infos3d.growedProbas[RMAP_FLAT] = 1.0 - maxObstacleProba;
	region->lastDate = rm->nbUpdates;
      }
      else {
	region->infos3d.growedProbas[RMAP_OBSTACLE] 
	  = region->infos3d.mergedProbas[RMAP_OBSTACLE];
	region->infos3d.growedProbas[RMAP_FLAT] 
	  = region->infos3d.mergedProbas[RMAP_FLAT];
      }
    }
  }
  
  return (RMAP_OK);
}

/**********************************************************************
 *
 * Updates a circular area specified in metric coordinates.
 *
 **********************************************************************/

int
rMap_updateCircularArea (REGION_MAP*	regionMap,
			 double		xCenter,
			 double		yCenter,
			 double		radius,
			 double		probaObst)
{
  int		iC, jC;
  double	radiusPixels;

  if ((probaObst < 0.0) || (probaObst > 1.0)) {
    printf ("The proba must belong to [0.0 ; 1.0]\n");
    return rMapError (ERR_RMAP_BAD_PARAMETER);
  }

  /* Convert metric coordinates in pixel coordinates */
  iC = (int) (((xCenter - regionMap->origin2corner.euler.euler[3]) / regionMap->scale) - 0.5);  
  jC = (int) (((yCenter - regionMap->origin2corner.euler.euler[4]) / regionMap->scale) - 0.5);  
  radiusPixels = radius / regionMap->scale;
    
  rMap_updateCircle (regionMap, iC, jC, radiusPixels, probaObst);
  
  /* Update the various global informations` on the regionMap */
  rMap_endUpdate (regionMap);
  
  return (RMAP_OK);
}

/**********************************************************************
 *
 * Updates a circle area specified in pixel coordinates
 *
 **********************************************************************/
void
rMap_updateCircle (REGION_MAP*	regionMap,
		   int		iC,
		   int		jC,
		   double	radius,
		   double	probaObst)
{
  int		i, j;
  int		imin, jmin, imax, jmax;
  int		currentDate;
  double	radius2;
  // Setup cell-scan domain limits
  imin = iC - ((int)radius) - 1;
  if (imin < 0) { imin = 0; }
  jmin = jC - ((int)radius) - 1;
  if (jmin < 0) { jmin = 0; }
  imax = iC + ((int)radius) + 1;
  if (imax >= regionMap->nblig) { imax = regionMap->nblig - 1; }
  jmax = jC + ((int)radius) + 1;
  if (jmax >= regionMap->nbcol) { jmax = regionMap->nbcol - 1; }
  // Work with squared pixel distances
  radius2 = radius * radius;
  currentDate = regionMap->nbUpdates + 1;

  /*
  for (i = 0; i < regionMap->nblig; i++) {
    for (j = 0; j < regionMap->nbcol; j++) {
  */
  	for (i = imin; i <= imax; i++) {
		for (j = jmin; j <= jmax; j++) {
			if ( ((j-jC)*(j-jC) + (i-iC)*(i-iC)) < radius2 ) {
				rMap_updateRegionWithProba (regionMap, i, j, currentDate, probaObst);

			}
		}
	}
	return;
}

/**********************************************************************
 *
 * Updates a rectangular area specified in pixel coordinates
 *
 **********************************************************************/
void
rMap_updateRectangle (REGION_MAP*	regionMap,
		      int		iMin,
		      int		iMax,
		      int		jMin,
		      int		jMax,
		      double		probaObst)
{
  int		i, j;
  int		currentDate;

  currentDate = regionMap->nbUpdates + 1;
  for (i = iMin; i <= iMax; i++) {
    for (j = jMin; j <= jMax; j++) {
      rMap_updateRegionWithProba (regionMap, i, j, currentDate, probaObst);
    }
  }
  return;
}

/**********************************************************************
 *
 * Updates a rectangular area specified in metric coordinates.
 *
 **********************************************************************/

int
rMap_updateRectangularArea (REGION_MAP*	regionMap,
			    int		xMin,
			    int		xMax,
			    int		yMin,
			    int		yMax,
			    double	probaObst)

{
  int		iMin, jMin, iMax, jMax;

  if ((probaObst < 0.0) || (probaObst > 1.0)) {
    printf ("The proba must belong to [0.0 ; 1.0]\n");
    return rMapError (ERR_RMAP_BAD_PARAMETER);
  }
    
  iMin = (xMin - regionMap->origin2corner.euler.euler[3]) / regionMap->scale;
  iMax = (xMax - regionMap->origin2corner.euler.euler[3]) / regionMap->scale;
  jMin = (yMin - regionMap->origin2corner.euler.euler[4]) / regionMap->scale;
  jMax = (yMax - regionMap->origin2corner.euler.euler[4]) / regionMap->scale;
  
  rMap_updateRectangle (regionMap, iMin, iMax, jMin, jMax, probaObst);
  
  /* Update the various global informations` on the regionMap */
  rMap_endUpdate (regionMap);
  
  return (RMAP_OK);
}

/**********************************************************************
 *
 * Updates a "linear area" (a line actually) specified in pixel coordinates
 *
 **********************************************************************/
int
rMap_updateLine (REGION_MAP*	regionMap,
		 int		i1,
		 int		j1,
		 int		i2,
		 int		j2,
		 double		probaObst)
{
  int		i, j;
  int		currentDate;

  /* The following variables are for the Bressenham algo */
  int xdelta;			/* The change in x coordinates */
  int ydelta;			/* The change in y coordinates */
  int xstep;			/* The change to make in the x coordinate in
				 * each step */
  int ystep;			/* The change to make in the y coordinate in
				 * each step */
  int change;			/* The amount that the x or y coordinate has
				 * changed */

  currentDate = regionMap->nbUpdates + 1;
  if ((i1 < 0) || (i1 >= regionMap->nblig)
      || (i2 < 0) || (i2 >= regionMap->nblig)
      || (j1 < 0) || (j1 >= regionMap->nbcol)
      || (j2 < 0) || (j2 >= regionMap->nbcol)) {
    printf ("The coordinates must be within the map boudaries\n");
    return rMapError (ERR_RMAP_BAD_PARAMETER);
  }

  xdelta = i2 - i1;		/* Calculate the change in x coordinates */
  ydelta = j2 - j1;		/* Calculate the change in y coordinates */
  if (xdelta < 0) {		/* The line will be drawn from right to left */
    xdelta = -xdelta;
    xstep = -1;
  } else			/* The line will be drawn from left to right */
    xstep = 1;
  if (ydelta < 0) {		/* The line will be drawn from bottom to top */
    ydelta = -ydelta;
    ystep = -1;
  } else			/* The line will be drawn from top to bottom */
    ystep = 1;
  if (xdelta > ydelta) {	/* x changes quicker than y */
    change = xdelta >> 1;	/* change set to twice the value of xdelta */
    while (i1 != i2) {	/* Draw until the terminating dot is reached */
      /* Update the (i1, j1) region */
      rMap_updateRegionWithProba (regionMap, i1, j1, currentDate, probaObst);

      i1 += xstep;	/* Update x coordinate */
      change += ydelta;	/* Update change */
      if (change > xdelta)
	/* If change is large enough to update the y coordinate */
	{
	  j1 += ystep;	/* Update the y coordinate */
	  change -= xdelta;	/* Reset change */
	}
    }
  } else {			/* y changes quicker than x */
    change = ydelta >> 1;	/* change set to twice the value of ydelta */
    while (j1 != j2) {	/* Draw until the terminating dot is reached */
      /* Update the (i1, j1) region */
      rMap_updateRegionWithProba (regionMap, i1, j1, currentDate, probaObst);
      
      j1 += ystep;	/* Update y coordinate */
      change += xdelta;	/* Update change */
      if (change > ydelta)
	/* If change is large enough to update the x coordinate */
	{
	  i1 += xstep;	/* Update the x coordinate */
	  change -= ydelta;	/* Reset change */
	}
    }
  }
  return (RMAP_OK);
}

/**********************************************************************
 *
 * Updates a "linear area" (a line actually) specified from a given position by
 * a distance and a width
 *
 **********************************************************************/
int
rMap_updateLineFromPosition (REGION_MAP*	regionMap,
			     double		x,
			     double		y,
			     double		theta,
			     double		distance,
			     double		length,
			     double		probaObst)
{
  double	xc, yc;		/* Center of the line */
  double	x1, y1;
  double	x2, y2;		/* Extremities of the line */
  int		i1, j1;
  int		i2, j2;

  xc = x + distance * cos(theta);
  yc = y + distance * sin(theta);
  x1 = xc - (length / 2) * sin(theta);
  y1 = yc + (length / 2) * cos(theta);
  x2 = xc + (length / 2) * sin(theta);
  y2 = yc - (length / 2) * cos(theta);

  i1 = lround ((x1 - regionMap->origin2corner.euler.euler[3]) /regionMap->scale);
  j1 = lround ((y1 - regionMap->origin2corner.euler.euler[4]) /regionMap->scale);
  i2 = lround ((x2 - regionMap->origin2corner.euler.euler[3]) /regionMap->scale);
  j2 = lround ((y2 - regionMap->origin2corner.euler.euler[4]) /regionMap->scale);

  return (rMap_updateLine (regionMap, i1, j1, i2, j2, probaObst));
}

/***********************************************************************
 *
 *
 *
 ***********************************************************************/

void 
rMap_updateRegionWithProba (REGION_MAP*	regionMap,
			    int		i,
			    int		j,
			    int		currentDate,
			    double	probaObst)

{
  RMAP_REGION*	region;

  region = &(regionMap->regions[i * regionMap->nbcol + j]);

  /* If this is a newly perceived region, update some global region and
     regionMap characteristics  */
  if (region->lastDate != currentDate) {
    rMap_updateStates (regionMap, region, i, j, currentDate);
    region->updateState = RMAP_REGION_INITIALLY_PERCEIVED;
  }
  else {
    /* The region has already been perceived -- update its lastDate */
    region->lastDate = currentDate;
  }
  
  /* Update its probabilities */
  region->infos3d.probaState = RMAP_PROBAS_COMPUTED;
  region->infos3d.newProbas[RMAP_FLAT] = 1.0 - probaObst;
  region->infos3d.mergedProbas[RMAP_FLAT] = 1.0 - probaObst;
  region->infos3d.newProbas[RMAP_OBSTACLE] = probaObst;
  region->infos3d.mergedProbas[RMAP_OBSTACLE] = probaObst;
  region->classificationType[CLASSIFICATION_3D] = 1;
  region->infos3d.classificationState = EXTERNALLY_CLASSIFIED;
  region->updateState = RMAP_REGION_INITIALLY_PERCEIVED;

  return;
}
